from requests import JSONDecodeError, request
from tenacity import retry, retry_if_exception, stop_after_attempt, wait_exponential

from chainalysis.util_functions.exceptions import (
    BadRequest,
    DataSolutionsAPIException,
)


def get_headers(api_key: str) -> dict:
    """
    Generate headers for an HTTP request.

    This function creates a dictionary of headers required for an HTTP request,
    including the API key for authorization and JSON content-type headers which
    indicate what sort of information will be sent with the request.

    :param api_key: The API key used for authenticating the request.
    :return: A dictionary containing the headers for the HTTP request.
    """

    return {
        "X-API-KEY": api_key,
        "Content-Type": "application/json",
        "Accept": "application/json",
    }


def retry_if_not_bad_request(exception):
    if isinstance(exception, BadRequest):
        return False  # Do not retry if it's a BadRequestException
    return True  # Retry for other exceptions


@retry(
    wait=wait_exponential(multiplier=1, min=4, max=10),
    retry=retry_if_exception(retry_if_not_bad_request),
    stop=stop_after_attempt(10),
)
def issue_request(
    url: str,
    api_key: str,
    params: dict = {},
    body: dict = {},
    method: str = "GET",
) -> dict:
    """
    Helper method to issue a request to the Data Solutions API.
    This method will automatically retry the request, and handle
    basic error checking.
    """

    # Encode params and body into a single variable data
    data = {**(params or {}), **(body or {})}

    response = request(
        method,
        url,
        headers=get_headers(api_key),
        json=data,
    )

    if response.status_code >= 300:
        try:  # Try to decode the json response of the error
            json_output = response.json()
            error_message = json_output.get("message")

        except JSONDecodeError:
            raise DataSolutionsAPIException(
                message="Unexpected response from the API - response was not JSON",
            )
        if response.status_code == 400:
            raise BadRequest(
                message=error_message,
            )
        raise DataSolutionsAPIException(
            message=error_message,
        )

    try:  # Try to decode the json response of the successful request
        response_json = response.json()
    except JSONDecodeError:
        raise DataSolutionsAPIException(
            message="Unexpected response from the API - response was not JSON",
        )

    return response_json
